/*  DCC_Monitor for LCD 4x20

  Based on DCC_Decoder Library Version 4 from http://www.mynabay.com/arduino
  Display DCC packets counts every 2 seconds (IDLE should be around 50%)
  Arduino Pin 2 is the DCC input. It drives interrupt 0.
  Arduino Pin A4 is SDA, Pin A5 is SLC (even on Nano !). To connect an LCD 20x4 display via I2C
  Don't forget I2C pull-up resistors (4,7K each) connected to + 5V
  Compilation on Arduino 1.5.x necessitate a minor change in the DCC_Decoder library:
  Replace "prog_char*" type by "char*"
  
  Display 1st line : Idle paquets count, valid data paquets count
  Display next lines : for each data type :
  - paquet count, DCC address, 
    - speed paquet : direction (< or >), speed value (0..127)
    - light function paquet (FL) : L0 (off) or L1 (on)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation.
  
  Copyright (c) 2014 Dominique Bultez. */

#define VERSION "1.0b3 D.BULTEZ 10.14"
#define D_USB // comment this line to cancel display to the computer's terminal

#include <DCC_Decoder.h>
#include <Wire.h> 
#include <LiquidCrystal_I2C.h>

LiquidCrystal_I2C lcd(0x27,20,4);  // set the LCD address to 0x27 for a 20 chars and 4 line display

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Defines and structures
//
#define kDCC_INTERRUPT            0

typedef struct
{
    int count;
    byte validBytes;
    byte data[6];
} DCCPacket;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// The dcc decoder object and global data
//
int gPacketCount = 0;
int gIdlePacketCount = 0;
int gLongestPreamble = 0;

DCCPacket gPackets[25];

static unsigned long lastMillis = millis();
    
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Packet handlers
//

// ALL packets are sent to the RawPacket handler. Returning true indicates that packet was handled. DCC library starts watching for 
// next preamble. Returning false and library continue parsing packet and finds another handler to call.
boolean RawPacket_Handler(byte byteCount, byte* packetBytes)
{
        // Bump global packet count
    ++gPacketCount;
    
    int thisPreamble = DCC.LastPreambleBitCount();
    if( thisPreamble > gLongestPreamble )
    {
        gLongestPreamble = thisPreamble;
    }
    
        // Walk table and look for a matching packet
    for( int i=0; i<(int)(sizeof(gPackets)/sizeof(gPackets[0])); ++i )
    {
        if( gPackets[i].validBytes )
        {
                // Not an empty slot. Does this slot match this packet? If so, bump count.
            if( gPackets[i].validBytes==byteCount )
            {
                char isPacket = true;
                for( int j=0; j<byteCount; j++)
                {
                    if( gPackets[i].data[j] != packetBytes[j] )
                    {
                        isPacket = false;
                        break;
                    } 
                }
                if( isPacket )
                {
                   gPackets[i].count++;
                   return false;
                }
            }
        }else{
                // Empty slot, just copy over data
            gPackets[i].count++;
            gPackets[i].validBytes = byteCount;
            for( int j=0; j<byteCount; j++)
            {
                gPackets[i].data[j] = packetBytes[j];
            }
            return false;
        }
    }    
    
    return false;
}

// Idle packets are sent here (unless handled in rawpacket handler). 
void IdlePacket_Handler(byte byteCount, byte* packetBytes)
{
    ++gIdlePacketCount;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Setup
//
void setup() 
{ 
   Serial.begin(115200);
   lcd.init();                      // initialize the lcd 
   lcd.backlight();

   lcd.print("DCC monitor Version ");
   lcd.print(VERSION);
   delay(1000);
   //lcd.clear();
   
   DCC.SetRawPacketHandler(RawPacket_Handler);   
   DCC.SetIdlePacketHandler(IdlePacket_Handler);
            
   DCC.SetupMonitor( kDCC_INTERRUPT );   
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DumpAndResetTable()
{
    char buffer60Bytes[60];
    int i=0;
    int j=0;
    
    #ifdef D_USB
    Serial.print("Total Packet Count: ");
    Serial.println(gPacketCount, DEC);
    #endif
    
    lcd.clear();
    //lcd.setCursor(0, L);    // 1e ligne
    lcd.print("Idle:");
    lcd.print(gIdlePacketCount, DEC);

    #ifdef D_USB
    Serial.print("Idle Packet Count:  ");
    Serial.println(gIdlePacketCount, DEC);
    #endif
    
    lcd.print(" Data:");
    lcd.print(gPacketCount - gIdlePacketCount, DEC);
    lcd.print("  ");    
    
    #ifdef D_USB
    Serial.print("Longest Preamble:  ");
    Serial.println(gLongestPreamble, DEC);    
    Serial.println("Count    Packet_Data");
    #endif
    
    lcd.setCursor(0, 1);    // 2e ligne
    for( i=0; i<(int)(sizeof(gPackets)/sizeof(gPackets[0])); ++i )
    {
        if( gPackets[i].validBytes > 0 )
        {
            if (gPackets[i].count != gIdlePacketCount)
            {
              #ifdef D_USB
              Serial.print(gPackets[i].count, DEC);                                
              if( gPackets[i].count < 10 )
              {
                Serial.print("        ");
              }else{
                if( gPackets[i].count < 100 )
                {
                    Serial.print("       ");
                }else{
                    Serial.print("      ");
                }
              }
              Serial.println( DCC.MakePacketString(buffer60Bytes, gPackets[i].validBytes, &gPackets[i].data[0]) );
              #endif
              
              if (gPackets[i].data[0] != 255)
              {
                lcd.print(gPackets[i].count, DEC);   // count
                lcd.print("@");                      // @
                lcd.print(gPackets[i].data[0], DEC); // DCC address
                switch (gPackets[i].data[1]) 
                {
                  case 0x3F:  // Advanced Operation Instruction : speed & direction
                  if (gPackets[i].data[2] > 127)
                  {
                    lcd.print(">");
                    lcd.print(gPackets[i].data[2] - 128); //  > speed
                  }else{
                    lcd.print("<");
                    lcd.print(gPackets[i].data[2]);       //  < speed
                  }
                  break;
                  case 0x90:  // Fonction Group One FL on
                  lcd.print("L1");
                  break;
                  case 0x80:  // Fonction Group One FL off
                  lcd.print("L0");
                  break;
                  default:
                  lcd.print(gPackets[i].data[1], HEX);
                  break;                  
                }
                lcd.print("   ");
                j++;
                if (j>5) j=0;
                switch (j)
                {
                  case 0:
                  lcd.setCursor(0, 1);    // 2e ligne
                  break;
                  case 1:
                  lcd.setCursor(10, 1);    // 2e ligne
                  break;
                  case 2:
                  lcd.setCursor(0, 2);    // 2e ligne
                  break;
                  case 3:
                  lcd.setCursor(10, 2);    // 2e ligne
                  break;
                  case 4:
                  lcd.setCursor(0, 3);    // 2e ligne
                  break;
                  case 5:
                  lcd.setCursor(10, 3);    // 2e ligne
                  break;
                }
              }
            }
        }
        gPackets[i].validBytes = 0;
        gPackets[i].count = 0;
    }
    #ifdef D_USB
    Serial.println("============================================");
    #endif
    
    gPacketCount = 0;
    gIdlePacketCount = 0;
    gLongestPreamble = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Main loop
//
void loop()
{
    DCC.loop();
    
    if( millis()-lastMillis > 2000 )
    {
        DumpAndResetTable();
        lastMillis = millis();
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

